@@ -31,6 +31,7 @@
 #include "executor/spi.h"
 #include "miscadmin.h"
 #include "port/atomics.h"
+#include "storage/spin.h"
 #include "utils/builtins.h"
 #include "utils/geo_decls.h"
 #include "utils/rel.h"
@@ -38,6 +39,34 @@
 #include "utils/memutils.h"
 
 
+#define EXPECT_TRUE(expr)	\
+	do { \
+		if (!(expr)) \
+			elog(ERROR, \
+				 "%s was unexpectedly false in file \"%s\" line %u", \
+				 #expr, __FILE__, __LINE__); \
+	} while (0)
+
+#define EXPECT_EQ_U32(result_expr, expected_expr)	\
+	do { \
+		uint32		result = (result_expr); \
+		uint32		expected = (expected_expr); \
+		if (result != expected) \
+			elog(ERROR, \
+				 "%s yielded %u, expected %s in file \"%s\" line %u", \
+				 #result_expr, result, #expected_expr, __FILE__, __LINE__); \
+	} while (0)
+
+#define EXPECT_EQ_U64(result_expr, expected_expr)	\
+	do { \
+		uint64		result = (result_expr); \
+		uint64		expected = (expected_expr); \
+		if (result != expected) \
+			elog(ERROR, \
+				 "%s yielded " UINT64_FORMAT ", expected %s in file \"%s\" line %u", \
+				 #result_expr, result, #expected_expr, __FILE__, __LINE__); \
+	} while (0)
+
 #define P_MAXDIG 12
 #define LDELIM			'('
 #define RDELIM			')'
@@ -904,27 +933,13 @@ test_atomic_flag(void)
 	pg_atomic_flag flag;
 
 	pg_atomic_init_flag(&flag);
-
-	if (!pg_atomic_unlocked_test_flag(&flag))
-		elog(ERROR, "flag: unexpectedly set");
-
-	if (!pg_atomic_test_set_flag(&flag))
-		elog(ERROR, "flag: couldn't set");
-
-	if (pg_atomic_unlocked_test_flag(&flag))
-		elog(ERROR, "flag: unexpectedly unset");
-
-	if (pg_atomic_test_set_flag(&flag))
-		elog(ERROR, "flag: set spuriously #2");
-
+	EXPECT_TRUE(pg_atomic_unlocked_test_flag(&flag));
+	EXPECT_TRUE(pg_atomic_test_set_flag(&flag));
+	EXPECT_TRUE(!pg_atomic_unlocked_test_flag(&flag));
+	EXPECT_TRUE(!pg_atomic_test_set_flag(&flag));
 	pg_atomic_clear_flag(&flag);
-
-	if (!pg_atomic_unlocked_test_flag(&flag))
-		elog(ERROR, "flag: unexpectedly set #2");
-
-	if (!pg_atomic_test_set_flag(&flag))
-		elog(ERROR, "flag: couldn't set");
-
+	EXPECT_TRUE(pg_atomic_unlocked_test_flag(&flag));
+	EXPECT_TRUE(pg_atomic_test_set_flag(&flag));
 	pg_atomic_clear_flag(&flag);
 }
 
@@ -936,61 +951,38 @@ test_atomic_uint32(void)
 	int			i;
 
 	pg_atomic_init_u32(&var, 0);
-
-	if (pg_atomic_read_u32(&var) != 0)
-		elog(ERROR, "atomic_read_u32() #1 wrong");
-
+	EXPECT_EQ_U32(pg_atomic_read_u32(&var), 0);
 	pg_atomic_write_u32(&var, 3);
-
-	if (pg_atomic_read_u32(&var) != 3)
-		elog(ERROR, "atomic_read_u32() #2 wrong");
-
-	if (pg_atomic_fetch_add_u32(&var, 1) != 3)
-		elog(ERROR, "atomic_fetch_add_u32() #1 wrong");
-
-	if (pg_atomic_fetch_sub_u32(&var, 1) != 4)
-		elog(ERROR, "atomic_fetch_sub_u32() #1 wrong");
-
-	if (pg_atomic_sub_fetch_u32(&var, 3) != 0)
-		elog(ERROR, "atomic_sub_fetch_u32() #1 wrong");
-
-	if (pg_atomic_add_fetch_u32(&var, 10) != 10)
-		elog(ERROR, "atomic_add_fetch_u32() #1 wrong");
-
-	if (pg_atomic_exchange_u32(&var, 5) != 10)
-		elog(ERROR, "pg_atomic_exchange_u32() #1 wrong");
-
-	if (pg_atomic_exchange_u32(&var, 0) != 5)
-		elog(ERROR, "pg_atomic_exchange_u32() #0 wrong");
+	EXPECT_EQ_U32(pg_atomic_read_u32(&var), 3);
+	EXPECT_EQ_U32(pg_atomic_fetch_add_u32(&var, pg_atomic_read_u32(&var) - 2),
+				  3);
+	EXPECT_EQ_U32(pg_atomic_fetch_sub_u32(&var, 1), 4);
+	EXPECT_EQ_U32(pg_atomic_sub_fetch_u32(&var, 3), 0);
+	EXPECT_EQ_U32(pg_atomic_add_fetch_u32(&var, 10), 10);
+	EXPECT_EQ_U32(pg_atomic_exchange_u32(&var, 5), 10);
+	EXPECT_EQ_U32(pg_atomic_exchange_u32(&var, 0), 5);
 
 	/* test around numerical limits */
-	if (pg_atomic_fetch_add_u32(&var, INT_MAX) != 0)
-		elog(ERROR, "pg_atomic_fetch_add_u32() #2 wrong");
-
-	if (pg_atomic_fetch_add_u32(&var, INT_MAX) != INT_MAX)
-		elog(ERROR, "pg_atomic_add_fetch_u32() #3 wrong");
-
+	EXPECT_EQ_U32(pg_atomic_fetch_add_u32(&var, INT_MAX), 0);
+	EXPECT_EQ_U32(pg_atomic_fetch_add_u32(&var, INT_MAX), INT_MAX);
+	pg_atomic_fetch_add_u32(&var, 2);	/* wrap to 0 */
+	EXPECT_EQ_U32(pg_atomic_fetch_add_u32(&var, PG_INT16_MAX), 0);
+	EXPECT_EQ_U32(pg_atomic_fetch_add_u32(&var, PG_INT16_MAX + 1),
+				  PG_INT16_MAX);
+	EXPECT_EQ_U32(pg_atomic_fetch_add_u32(&var, PG_INT16_MIN),
+				  2 * PG_INT16_MAX + 1);
+	EXPECT_EQ_U32(pg_atomic_fetch_add_u32(&var, PG_INT16_MIN - 1),
+				  PG_INT16_MAX);
 	pg_atomic_fetch_add_u32(&var, 1);	/* top up to UINT_MAX */
-
-	if (pg_atomic_read_u32(&var) != UINT_MAX)
-		elog(ERROR, "atomic_read_u32() #2 wrong");
-
-	if (pg_atomic_fetch_sub_u32(&var, INT_MAX) != UINT_MAX)
-		elog(ERROR, "pg_atomic_fetch_sub_u32() #2 wrong");
-
-	if (pg_atomic_read_u32(&var) != (uint32) INT_MAX + 1)
-		elog(ERROR, "atomic_read_u32() #3 wrong: %u", pg_atomic_read_u32(&var));
-
-	expected = pg_atomic_sub_fetch_u32(&var, INT_MAX);
-	if (expected != 1)
-		elog(ERROR, "pg_atomic_sub_fetch_u32() #3 wrong: %u", expected);
-
+	EXPECT_EQ_U32(pg_atomic_read_u32(&var), UINT_MAX);
+	EXPECT_EQ_U32(pg_atomic_fetch_sub_u32(&var, INT_MAX), UINT_MAX);
+	EXPECT_EQ_U32(pg_atomic_read_u32(&var), (uint32) INT_MAX + 1);
+	EXPECT_EQ_U32(pg_atomic_sub_fetch_u32(&var, INT_MAX), 1);
 	pg_atomic_sub_fetch_u32(&var, 1);
 
 	/* fail exchange because of old expected */
 	expected = 10;
-	if (pg_atomic_compare_exchange_u32(&var, &expected, 1))
-		elog(ERROR, "atomic_compare_exchange_u32() changed value spuriously");
+	EXPECT_TRUE(!pg_atomic_compare_exchange_u32(&var, &expected, 1));
 
 	/* CAS is allowed to fail due to interrupts, try a couple of times */
 	for (i = 0; i < 1000; i++)
@@ -1001,31 +993,18 @@ test_atomic_uint32(void)
 	}
 	if (i == 1000)
 		elog(ERROR, "atomic_compare_exchange_u32() never succeeded");
-	if (pg_atomic_read_u32(&var) != 1)
-		elog(ERROR, "atomic_compare_exchange_u32() didn't set value properly");
-
+	EXPECT_EQ_U32(pg_atomic_read_u32(&var), 1);
 	pg_atomic_write_u32(&var, 0);
 
 	/* try setting flagbits */
-	if (pg_atomic_fetch_or_u32(&var, 1) & 1)
-		elog(ERROR, "pg_atomic_fetch_or_u32() #1 wrong");
-
-	if (!(pg_atomic_fetch_or_u32(&var, 2) & 1))
-		elog(ERROR, "pg_atomic_fetch_or_u32() #2 wrong");
-
-	if (pg_atomic_read_u32(&var) != 3)
-		elog(ERROR, "invalid result after pg_atomic_fetch_or_u32()");
-
+	EXPECT_TRUE(!(pg_atomic_fetch_or_u32(&var, 1) & 1));
+	EXPECT_TRUE(pg_atomic_fetch_or_u32(&var, 2) & 1);
+	EXPECT_EQ_U32(pg_atomic_read_u32(&var), 3);
 	/* try clearing flagbits */
-	if ((pg_atomic_fetch_and_u32(&var, ~2) & 3) != 3)
-		elog(ERROR, "pg_atomic_fetch_and_u32() #1 wrong");
-
-	if (pg_atomic_fetch_and_u32(&var, ~1) != 1)
-		elog(ERROR, "pg_atomic_fetch_and_u32() #2 wrong: is %u",
-			 pg_atomic_read_u32(&var));
+	EXPECT_EQ_U32(pg_atomic_fetch_and_u32(&var, ~2) & 3, 3);
+	EXPECT_EQ_U32(pg_atomic_fetch_and_u32(&var, ~1), 1);
 	/* no bits set anymore */
-	if (pg_atomic_fetch_and_u32(&var, ~0) != 0)
-		elog(ERROR, "pg_atomic_fetch_and_u32() #3 wrong");
+	EXPECT_EQ_U32(pg_atomic_fetch_and_u32(&var, ~0), 0);
 }
 
 #ifdef PG_HAVE_ATOMIC_U64_SUPPORT
@@ -1037,37 +1016,20 @@ test_atomic_uint64(void)
 	int			i;
 
 	pg_atomic_init_u64(&var, 0);
-
-	if (pg_atomic_read_u64(&var) != 0)
-		elog(ERROR, "atomic_read_u64() #1 wrong");
-
+	EXPECT_EQ_U64(pg_atomic_read_u64(&var), 0);
 	pg_atomic_write_u64(&var, 3);
-
-	if (pg_atomic_read_u64(&var) != 3)
-		elog(ERROR, "atomic_read_u64() #2 wrong");
-
-	if (pg_atomic_fetch_add_u64(&var, 1) != 3)
-		elog(ERROR, "atomic_fetch_add_u64() #1 wrong");
-
-	if (pg_atomic_fetch_sub_u64(&var, 1) != 4)
-		elog(ERROR, "atomic_fetch_sub_u64() #1 wrong");
-
-	if (pg_atomic_sub_fetch_u64(&var, 3) != 0)
-		elog(ERROR, "atomic_sub_fetch_u64() #1 wrong");
-
-	if (pg_atomic_add_fetch_u64(&var, 10) != 10)
-		elog(ERROR, "atomic_add_fetch_u64() #1 wrong");
-
-	if (pg_atomic_exchange_u64(&var, 5) != 10)
-		elog(ERROR, "pg_atomic_exchange_u64() #1 wrong");
-
-	if (pg_atomic_exchange_u64(&var, 0) != 5)
-		elog(ERROR, "pg_atomic_exchange_u64() #0 wrong");
+	EXPECT_EQ_U64(pg_atomic_read_u64(&var), 3);
+	EXPECT_EQ_U64(pg_atomic_fetch_add_u64(&var, pg_atomic_read_u64(&var) - 2),
+				  3);
+	EXPECT_EQ_U64(pg_atomic_fetch_sub_u64(&var, 1), 4);
+	EXPECT_EQ_U64(pg_atomic_sub_fetch_u64(&var, 3), 0);
+	EXPECT_EQ_U64(pg_atomic_add_fetch_u64(&var, 10), 10);
+	EXPECT_EQ_U64(pg_atomic_exchange_u64(&var, 5), 10);
+	EXPECT_EQ_U64(pg_atomic_exchange_u64(&var, 0), 5);
 
 	/* fail exchange because of old expected */
 	expected = 10;
-	if (pg_atomic_compare_exchange_u64(&var, &expected, 1))
-		elog(ERROR, "atomic_compare_exchange_u64() changed value spuriously");
+	EXPECT_TRUE(!pg_atomic_compare_exchange_u64(&var, &expected, 1));
 
 	/* CAS is allowed to fail due to interrupts, try a couple of times */
 	for (i = 0; i < 100; i++)
@@ -1078,34 +1040,172 @@ test_atomic_uint64(void)
 	}
 	if (i == 100)
 		elog(ERROR, "atomic_compare_exchange_u64() never succeeded");
-	if (pg_atomic_read_u64(&var) != 1)
-		elog(ERROR, "atomic_compare_exchange_u64() didn't set value properly");
+	EXPECT_EQ_U64(pg_atomic_read_u64(&var), 1);
 
 	pg_atomic_write_u64(&var, 0);
 
 	/* try setting flagbits */
-	if (pg_atomic_fetch_or_u64(&var, 1) & 1)
-		elog(ERROR, "pg_atomic_fetch_or_u64() #1 wrong");
-
-	if (!(pg_atomic_fetch_or_u64(&var, 2) & 1))
-		elog(ERROR, "pg_atomic_fetch_or_u64() #2 wrong");
-
-	if (pg_atomic_read_u64(&var) != 3)
-		elog(ERROR, "invalid result after pg_atomic_fetch_or_u64()");
-
+	EXPECT_TRUE(!(pg_atomic_fetch_or_u64(&var, 1) & 1));
+	EXPECT_TRUE(pg_atomic_fetch_or_u64(&var, 2) & 1);
+	EXPECT_EQ_U64(pg_atomic_read_u64(&var), 3);
 	/* try clearing flagbits */
-	if ((pg_atomic_fetch_and_u64(&var, ~2) & 3) != 3)
-		elog(ERROR, "pg_atomic_fetch_and_u64() #1 wrong");
-
-	if (pg_atomic_fetch_and_u64(&var, ~1) != 1)
-		elog(ERROR, "pg_atomic_fetch_and_u64() #2 wrong: is " UINT64_FORMAT,
-			 pg_atomic_read_u64(&var));
+	EXPECT_EQ_U64((pg_atomic_fetch_and_u64(&var, ~2) & 3), 3);
+	EXPECT_EQ_U64(pg_atomic_fetch_and_u64(&var, ~1), 1);
 	/* no bits set anymore */
-	if (pg_atomic_fetch_and_u64(&var, ~0) != 0)
-		elog(ERROR, "pg_atomic_fetch_and_u64() #3 wrong");
+	EXPECT_EQ_U64(pg_atomic_fetch_and_u64(&var, ~0), 0);
 }
 #endif   /* PG_HAVE_ATOMIC_U64_SUPPORT */
 
+/*
+ * Perform, fairly minimal, testing of the spinlock implementation.
+ *
+ * It's likely worth expanding these to actually test concurrency etc, but
+ * having some regularly run tests is better than none.
+ */
+static void
+test_spinlock(void)
+{
+	/*
+	 * Basic tests for spinlocks, as well as the underlying operations.
+	 *
+	 * We embed the spinlock in a struct with other members to test that the
+	 * spinlock operations don't perform too wide writes.
+	 */
+	{
+		struct test_lock_struct
+		{
+			char		data_before[4];
+			slock_t		lock;
+			char		data_after[4];
+		} struct_w_lock;
+
+		memcpy(struct_w_lock.data_before, "abcd", 4);
+		memcpy(struct_w_lock.data_after, "ef12", 4);
+
+		/* test basic operations via the SpinLock* API */
+		SpinLockInit(&struct_w_lock.lock);
+		SpinLockAcquire(&struct_w_lock.lock);
+		SpinLockRelease(&struct_w_lock.lock);
+
+		/* test basic operations via underlying S_* API */
+		S_INIT_LOCK(&struct_w_lock.lock);
+		S_LOCK(&struct_w_lock.lock);
+		S_UNLOCK(&struct_w_lock.lock);
+
+		/* and that "contended" acquisition works */
+		s_lock(&struct_w_lock.lock, "testfile", 17, "testfunc");
+		S_UNLOCK(&struct_w_lock.lock);
+
+		/*
+		 * Check, using TAS directly, that a single spin cycle doesn't block
+		 * when acquiring an already acquired lock.
+		 */
+#ifdef TAS
+		S_LOCK(&struct_w_lock.lock);
+
+		if (!TAS(&struct_w_lock.lock))
+			elog(ERROR, "acquired already held spinlock");
+
+#ifdef TAS_SPIN
+		if (!TAS_SPIN(&struct_w_lock.lock))
+			elog(ERROR, "acquired already held spinlock");
+#endif							/* defined(TAS_SPIN) */
+
+		S_UNLOCK(&struct_w_lock.lock);
+#endif							/* defined(TAS) */
+
+		/*
+		 * Verify that after all of this the non-lock contents are still
+		 * correct.
+		 */
+		if (memcmp(struct_w_lock.data_before, "abcd", 4) != 0)
+			elog(ERROR, "padding before spinlock modified");
+		if (memcmp(struct_w_lock.data_after, "ef12", 4) != 0)
+			elog(ERROR, "padding after spinlock modified");
+	}
+
+	/*
+	 * Ensure that allocating more than INT32_MAX emulated spinlocks
+	 * works. That's interesting because the spinlock emulation uses a 32bit
+	 * integer to map spinlocks onto semaphores. There've been bugs...
+	 */
+#ifndef HAVE_SPINLOCKS
+	{
+		uint32	i;
+
+		/*
+		 * Initialize enough spinlocks to advance counter close to
+		 * wraparound. It's too expensive to perform acquire/release for each,
+		 * as those may be syscalls when the spinlock emulation is used (and
+		 * even just atomic TAS would be expensive).
+		 */
+		for (i = 0; i < INT32_MAX - 100000; i++)
+		{
+			slock_t lock;
+
+			SpinLockInit(&lock);
+		}
+
+		for (i = 0; i < 200000; i++)
+		{
+			slock_t lock;
+
+			SpinLockInit(&lock);
+
+			SpinLockAcquire(&lock);
+			SpinLockRelease(&lock);
+			SpinLockAcquire(&lock);
+			SpinLockRelease(&lock);
+		}
+	}
+#endif
+}
+
+/*
+ * Verify that performing atomic ops inside a spinlock isn't a
+ * problem. Realistically that's only going to be a problem when both
+ * --disable-spinlocks and --disable-atomics are used, but it's cheap enough
+ * to just always test.
+ *
+ * The test works by initializing enough atomics that we'd conflict if there
+ * were an overlap between a spinlock and an atomic by holding a spinlock
+ * while manipulating more than NUM_SPINLOCK_SEMAPHORES atomics.
+ *
+ * NUM_TEST_ATOMICS doesn't really need to be more than
+ * NUM_SPINLOCK_SEMAPHORES, but it seems better to test a bit more
+ * extensively.
+ */
+static void
+test_atomic_spin_nest(void)
+{
+	slock_t lock;
+#define NUM_TEST_ATOMICS (NUM_SPINLOCK_SEMAPHORES + NUM_ATOMICS_SEMAPHORES + 27)
+	pg_atomic_uint32 atomics32[NUM_TEST_ATOMICS];
+	int		i;
+
+	SpinLockInit(&lock);
+
+	for (i = 0; i < NUM_TEST_ATOMICS; i++)
+	{
+		pg_atomic_init_u32(&atomics32[i], 0);
+	}
+
+	/* just so it's not all zeroes */
+	for (i = 0; i < NUM_TEST_ATOMICS; i++)
+	{
+		EXPECT_EQ_U32(pg_atomic_fetch_add_u32(&atomics32[i], i), 0);
+	}
+
+	/* test whether we can do atomic op with lock held */
+	SpinLockAcquire(&lock);
+	for (i = 0; i < NUM_TEST_ATOMICS; i++)
+	{
+		EXPECT_EQ_U32(pg_atomic_fetch_sub_u32(&atomics32[i], i), i);
+		EXPECT_EQ_U32(pg_atomic_read_u32(&atomics32[i]), 0);
+	}
+	SpinLockRelease(&lock);
+}
+#undef NUM_TEST_ATOMICS
 
 PG_FUNCTION_INFO_V1(test_atomic_ops);
 Datum
@@ -1119,5 +1219,13 @@ test_atomic_ops(PG_FUNCTION_ARGS)
 	test_atomic_uint64();
 #endif
 
+	/*
+	 * Arguably this shouldn't be tested as part of this function, but it's
+	 * closely enough related that that seems ok for now.
+	 */
+	test_spinlock();
+
+	test_atomic_spin_nest();
+
 	PG_RETURN_BOOL(true);
 }
